from esphome import automation
from esphome.automation import register_action, register_condition
import esphome.codegen as cg
from esphome.components import media_player, micro_wake_word, microphone, speaker
import esphome.config_validation as cv
from esphome.const import (
    CONF_ID,
    CONF_MEDIA_PLAYER,
    CONF_MICROPHONE,
    CONF_ON_CLIENT_CONNECTED,
    CONF_ON_CLIENT_DISCONNECTED,
    CONF_ON_ERROR,
    CONF_ON_IDLE,
    CONF_SPEAKER,
)

AUTO_LOAD = ["socket"]
DEPENDENCIES = ["api", "microphone"]

CODEOWNERS = ["@jesserockz", "@kahrendt"]

CONF_ON_END = "on_end"
CONF_ON_INTENT_END = "on_intent_end"
CONF_ON_INTENT_PROGRESS = "on_intent_progress"
CONF_ON_INTENT_START = "on_intent_start"
CONF_ON_LISTENING = "on_listening"
CONF_ON_START = "on_start"
CONF_ON_STT_END = "on_stt_end"
CONF_ON_STT_VAD_END = "on_stt_vad_end"
CONF_ON_STT_VAD_START = "on_stt_vad_start"
CONF_ON_TTS_END = "on_tts_end"
CONF_ON_TTS_START = "on_tts_start"
CONF_ON_TTS_STREAM_START = "on_tts_stream_start"
CONF_ON_TTS_STREAM_END = "on_tts_stream_end"
CONF_ON_WAKE_WORD_DETECTED = "on_wake_word_detected"

CONF_SILENCE_DETECTION = "silence_detection"
CONF_USE_WAKE_WORD = "use_wake_word"
CONF_VAD_THRESHOLD = "vad_threshold"

CONF_AUTO_GAIN = "auto_gain"
CONF_NOISE_SUPPRESSION_LEVEL = "noise_suppression_level"
CONF_VOLUME_MULTIPLIER = "volume_multiplier"

CONF_MICRO_WAKE_WORD = "micro_wake_word"
CONF_WAKE_WORD = "wake_word"

CONF_CONVERSATION_TIMEOUT = "conversation_timeout"

CONF_ON_TIMER_STARTED = "on_timer_started"
CONF_ON_TIMER_UPDATED = "on_timer_updated"
CONF_ON_TIMER_CANCELLED = "on_timer_cancelled"
CONF_ON_TIMER_FINISHED = "on_timer_finished"
CONF_ON_TIMER_TICK = "on_timer_tick"


voice_assistant_ns = cg.esphome_ns.namespace("voice_assistant")
VoiceAssistant = voice_assistant_ns.class_("VoiceAssistant", cg.Component)

StartAction = voice_assistant_ns.class_(
    "StartAction", automation.Action, cg.Parented.template(VoiceAssistant)
)
StartContinuousAction = voice_assistant_ns.class_(
    "StartContinuousAction", automation.Action, cg.Parented.template(VoiceAssistant)
)
StopAction = voice_assistant_ns.class_(
    "StopAction", automation.Action, cg.Parented.template(VoiceAssistant)
)
IsRunningCondition = voice_assistant_ns.class_(
    "IsRunningCondition", automation.Condition, cg.Parented.template(VoiceAssistant)
)
ConnectedCondition = voice_assistant_ns.class_(
    "ConnectedCondition", automation.Condition, cg.Parented.template(VoiceAssistant)
)

Timer = voice_assistant_ns.struct("Timer")


def tts_stream_validate(config):
    if CONF_SPEAKER not in config and (
        CONF_ON_TTS_STREAM_START in config or CONF_ON_TTS_STREAM_END in config
    ):
        raise cv.Invalid(
            f"{CONF_SPEAKER} is required when using {CONF_ON_TTS_STREAM_START} and/or {CONF_ON_TTS_STREAM_END}"
        )
    return config


CONFIG_SCHEMA = cv.All(
    cv.Schema(
        {
            cv.GenerateID(): cv.declare_id(VoiceAssistant),
            cv.Optional(
                CONF_MICROPHONE, default={}
            ): microphone.microphone_source_schema(
                min_bits_per_sample=16,
                max_bits_per_sample=16,
                min_channels=1,
                max_channels=1,
            ),
            cv.Exclusive(CONF_MEDIA_PLAYER, "output"): cv.use_id(
                media_player.MediaPlayer
            ),
            cv.Exclusive(CONF_SPEAKER, "output"): cv.use_id(speaker.Speaker),
            cv.Optional(CONF_USE_WAKE_WORD, default=False): cv.boolean,
            cv.Optional(CONF_MICRO_WAKE_WORD): cv.use_id(micro_wake_word.MicroWakeWord),
            cv.Optional(CONF_VAD_THRESHOLD): cv.invalid(
                "VAD threshold is no longer supported, as it requires the deprecated esp_adf external component. Use an i2s_audio microphone/speaker instead. Additionally, you may need to configure the audio_adc and audio_dac components depending on your hardware."
            ),
            cv.Optional(CONF_NOISE_SUPPRESSION_LEVEL, default=0): cv.int_range(0, 4),
            cv.Optional(CONF_AUTO_GAIN, default="0dBFS"): cv.All(
                cv.float_with_unit("decibel full scale", "(dBFS|dbfs|DBFS)"),
                cv.int_range(0, 31),
            ),
            cv.Optional(
                CONF_CONVERSATION_TIMEOUT, default="300s"
            ): cv.positive_time_period_milliseconds,
            cv.Optional(CONF_VOLUME_MULTIPLIER, default=1.0): cv.float_range(
                min=0.0, min_included=False
            ),
            cv.Optional(CONF_ON_LISTENING): automation.validate_automation(single=True),
            cv.Optional(CONF_ON_START): automation.validate_automation(single=True),
            cv.Optional(CONF_ON_WAKE_WORD_DETECTED): automation.validate_automation(
                single=True
            ),
            cv.Optional(CONF_ON_STT_END): automation.validate_automation(single=True),
            cv.Optional(CONF_ON_TTS_START): automation.validate_automation(single=True),
            cv.Optional(CONF_ON_TTS_END): automation.validate_automation(single=True),
            cv.Optional(CONF_ON_END): automation.validate_automation(single=True),
            cv.Optional(CONF_ON_ERROR): automation.validate_automation(single=True),
            cv.Optional(CONF_ON_CLIENT_CONNECTED): automation.validate_automation(
                single=True
            ),
            cv.Optional(CONF_ON_CLIENT_DISCONNECTED): automation.validate_automation(
                single=True
            ),
            cv.Optional(CONF_ON_INTENT_START): automation.validate_automation(
                single=True
            ),
            cv.Optional(CONF_ON_INTENT_PROGRESS): automation.validate_automation(
                single=True
            ),
            cv.Optional(CONF_ON_INTENT_END): automation.validate_automation(
                single=True
            ),
            cv.Optional(CONF_ON_STT_VAD_START): automation.validate_automation(
                single=True
            ),
            cv.Optional(CONF_ON_STT_VAD_END): automation.validate_automation(
                single=True
            ),
            cv.Optional(CONF_ON_TTS_STREAM_START): automation.validate_automation(
                single=True
            ),
            cv.Optional(CONF_ON_TTS_STREAM_END): automation.validate_automation(
                single=True
            ),
            cv.Optional(CONF_ON_IDLE): automation.validate_automation(single=True),
            cv.Optional(CONF_ON_TIMER_STARTED): automation.validate_automation(
                single=True
            ),
            cv.Optional(CONF_ON_TIMER_UPDATED): automation.validate_automation(
                single=True
            ),
            cv.Optional(CONF_ON_TIMER_CANCELLED): automation.validate_automation(
                single=True
            ),
            cv.Optional(CONF_ON_TIMER_FINISHED): automation.validate_automation(
                single=True
            ),
            cv.Optional(CONF_ON_TIMER_TICK): automation.validate_automation(
                single=True
            ),
        }
    ).extend(cv.COMPONENT_SCHEMA),
    tts_stream_validate,
)

FINAL_VALIDATE_SCHEMA = cv.All(
    cv.Schema(
        {
            cv.Optional(
                CONF_MICROPHONE
            ): microphone.final_validate_microphone_source_schema(
                "voice_assistant", sample_rate=16000
            ),
        },
        extra=cv.ALLOW_EXTRA,
    ),
)


async def to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    await cg.register_component(var, config)

    mic_source = await microphone.microphone_source_to_code(config[CONF_MICROPHONE])
    cg.add(var.set_microphone_source(mic_source))

    if CONF_MICRO_WAKE_WORD in config:
        mww = await cg.get_variable(config[CONF_MICRO_WAKE_WORD])
        cg.add(var.set_micro_wake_word(mww))

    if CONF_MEDIA_PLAYER in config:
        mp = await cg.get_variable(config[CONF_MEDIA_PLAYER])
        cg.add(var.set_media_player(mp))

    if CONF_SPEAKER in config:
        spkr = await cg.get_variable(config[CONF_SPEAKER])
        cg.add(var.set_speaker(spkr))

    cg.add(var.set_use_wake_word(config[CONF_USE_WAKE_WORD]))

    if (vad_threshold := config.get(CONF_VAD_THRESHOLD)) is not None:
        cg.add(var.set_vad_threshold(vad_threshold))

    cg.add(var.set_noise_suppression_level(config[CONF_NOISE_SUPPRESSION_LEVEL]))
    cg.add(var.set_auto_gain(config[CONF_AUTO_GAIN]))
    cg.add(var.set_volume_multiplier(config[CONF_VOLUME_MULTIPLIER]))
    cg.add(var.set_conversation_timeout(config[CONF_CONVERSATION_TIMEOUT]))

    if CONF_ON_LISTENING in config:
        await automation.build_automation(
            var.get_listening_trigger(), [], config[CONF_ON_LISTENING]
        )

    if CONF_ON_START in config:
        await automation.build_automation(
            var.get_start_trigger(), [], config[CONF_ON_START]
        )

    if CONF_ON_WAKE_WORD_DETECTED in config:
        await automation.build_automation(
            var.get_wake_word_detected_trigger(),
            [],
            config[CONF_ON_WAKE_WORD_DETECTED],
        )

    if CONF_ON_STT_END in config:
        await automation.build_automation(
            var.get_stt_end_trigger(), [(cg.std_string, "x")], config[CONF_ON_STT_END]
        )

    if CONF_ON_TTS_START in config:
        await automation.build_automation(
            var.get_tts_start_trigger(),
            [(cg.std_string, "x")],
            config[CONF_ON_TTS_START],
        )

    if CONF_ON_TTS_END in config:
        await automation.build_automation(
            var.get_tts_end_trigger(), [(cg.std_string, "x")], config[CONF_ON_TTS_END]
        )

    if CONF_ON_END in config:
        await automation.build_automation(
            var.get_end_trigger(), [], config[CONF_ON_END]
        )

    if CONF_ON_ERROR in config:
        await automation.build_automation(
            var.get_error_trigger(),
            [(cg.std_string, "code"), (cg.std_string, "message")],
            config[CONF_ON_ERROR],
        )

    if CONF_ON_CLIENT_CONNECTED in config:
        await automation.build_automation(
            var.get_client_connected_trigger(),
            [],
            config[CONF_ON_CLIENT_CONNECTED],
        )

    if CONF_ON_CLIENT_DISCONNECTED in config:
        await automation.build_automation(
            var.get_client_disconnected_trigger(),
            [],
            config[CONF_ON_CLIENT_DISCONNECTED],
        )

    if CONF_ON_INTENT_START in config:
        await automation.build_automation(
            var.get_intent_start_trigger(),
            [],
            config[CONF_ON_INTENT_START],
        )

    if CONF_ON_INTENT_PROGRESS in config:
        await automation.build_automation(
            var.get_intent_progress_trigger(),
            [(cg.std_string, "x")],
            config[CONF_ON_INTENT_PROGRESS],
        )

    if CONF_ON_INTENT_END in config:
        await automation.build_automation(
            var.get_intent_end_trigger(),
            [],
            config[CONF_ON_INTENT_END],
        )

    if CONF_ON_STT_VAD_START in config:
        await automation.build_automation(
            var.get_stt_vad_start_trigger(),
            [],
            config[CONF_ON_STT_VAD_START],
        )

    if CONF_ON_STT_VAD_END in config:
        await automation.build_automation(
            var.get_stt_vad_end_trigger(),
            [],
            config[CONF_ON_STT_VAD_END],
        )

    if CONF_ON_TTS_STREAM_START in config:
        await automation.build_automation(
            var.get_tts_stream_start_trigger(),
            [],
            config[CONF_ON_TTS_STREAM_START],
        )

    if CONF_ON_TTS_STREAM_END in config:
        await automation.build_automation(
            var.get_tts_stream_end_trigger(),
            [],
            config[CONF_ON_TTS_STREAM_END],
        )

    if CONF_ON_IDLE in config:
        await automation.build_automation(
            var.get_idle_trigger(),
            [],
            config[CONF_ON_IDLE],
        )

    has_timers = False
    if on_timer_started := config.get(CONF_ON_TIMER_STARTED):
        await automation.build_automation(
            var.get_timer_started_trigger(),
            [(Timer, "timer")],
            on_timer_started,
        )
        has_timers = True

    if on_timer_updated := config.get(CONF_ON_TIMER_UPDATED):
        await automation.build_automation(
            var.get_timer_updated_trigger(),
            [(Timer, "timer")],
            on_timer_updated,
        )
        has_timers = True

    if on_timer_cancelled := config.get(CONF_ON_TIMER_CANCELLED):
        await automation.build_automation(
            var.get_timer_cancelled_trigger(),
            [(Timer, "timer")],
            on_timer_cancelled,
        )
        has_timers = True

    if on_timer_finished := config.get(CONF_ON_TIMER_FINISHED):
        await automation.build_automation(
            var.get_timer_finished_trigger(),
            [(Timer, "timer")],
            on_timer_finished,
        )
        has_timers = True

    if on_timer_tick := config.get(CONF_ON_TIMER_TICK):
        await automation.build_automation(
            var.get_timer_tick_trigger(),
            [(cg.std_vector.template(Timer), "timers")],
            on_timer_tick,
        )
        has_timers = True

    cg.add(var.set_has_timers(has_timers))

    cg.add_define("USE_VOICE_ASSISTANT")


VOICE_ASSISTANT_ACTION_SCHEMA = cv.Schema({cv.GenerateID(): cv.use_id(VoiceAssistant)})


@register_action(
    "voice_assistant.start_continuous",
    StartContinuousAction,
    VOICE_ASSISTANT_ACTION_SCHEMA,
)
@register_action(
    "voice_assistant.start",
    StartAction,
    VOICE_ASSISTANT_ACTION_SCHEMA.extend(
        {
            cv.Optional(CONF_SILENCE_DETECTION, default=True): cv.boolean,
            cv.Optional(CONF_WAKE_WORD): cv.templatable(cv.string),
        }
    ),
)
async def voice_assistant_listen_to_code(config, action_id, template_arg, args):
    var = cg.new_Pvariable(action_id, template_arg)
    await cg.register_parented(var, config[CONF_ID])
    if CONF_SILENCE_DETECTION in config:
        cg.add(var.set_silence_detection(config[CONF_SILENCE_DETECTION]))
    if wake_word := config.get(CONF_WAKE_WORD):
        templ = await cg.templatable(wake_word, args, cg.std_string)
        cg.add(var.set_wake_word(templ))
    return var


@register_action("voice_assistant.stop", StopAction, VOICE_ASSISTANT_ACTION_SCHEMA)
async def voice_assistant_stop_to_code(config, action_id, template_arg, args):
    var = cg.new_Pvariable(action_id, template_arg)
    await cg.register_parented(var, config[CONF_ID])
    return var


@register_condition(
    "voice_assistant.is_running", IsRunningCondition, VOICE_ASSISTANT_ACTION_SCHEMA
)
async def voice_assistant_is_running_to_code(config, condition_id, template_arg, args):
    var = cg.new_Pvariable(condition_id, template_arg)
    await cg.register_parented(var, config[CONF_ID])
    return var


@register_condition(
    "voice_assistant.connected", ConnectedCondition, VOICE_ASSISTANT_ACTION_SCHEMA
)
async def voice_assistant_connected_to_code(config, condition_id, template_arg, args):
    var = cg.new_Pvariable(condition_id, template_arg)
    await cg.register_parented(var, config[CONF_ID])
    return var



from esphome import automation
from esphome.automation import maybe_simple_id
import esphome.codegen as cg
from esphome.components import i2c, sensor
import esphome.config_validation as cv
from esphome.const import CONF_GAIN, CONF_ID, ICON_SCALE, STATE_CLASS_MEASUREMENT

CODEOWNERS = ["@cujomalainey"]
DEPENDENCIES = ["i2c"]

CONF_GAIN_CALIBRATION = "gain_calibration"
CONF_OFFSET_CALIBRATION = "offset_calibration"
CONF_LDO_VOLTAGE = "ldo_voltage"
CONF_SAMPLES_PER_SECOND = "samples_per_second"

nau7802_ns = cg.esphome_ns.namespace("nau7802")
NAU7802Sensor = nau7802_ns.class_(
    "NAU7802Sensor", sensor.Sensor, cg.PollingComponent, i2c.I2CDevice
)
NAU7802CalbrateExternalOffsetAction = nau7802_ns.class_(
    "NAU7802CalbrateExternalOffsetAction",
    automation.Action,
    cg.Parented.template(NAU7802Sensor),
)
NAU7802CalbrateInternalOffsetAction = nau7802_ns.class_(
    "NAU7802CalbrateInternalOffsetAction",
    automation.Action,
    cg.Parented.template(NAU7802Sensor),
)
NAU7802CalbrateGainAction = nau7802_ns.class_(
    "NAU7802CalbrateGainAction", automation.Action, cg.Parented.template(NAU7802Sensor)
)

NAU7802Gain = nau7802_ns.enum("NAU7802Gain")
GAINS = {
    128: NAU7802Gain.NAU7802_GAIN_128,
    64: NAU7802Gain.NAU7802_GAIN_64,
    32: NAU7802Gain.NAU7802_GAIN_32,
    16: NAU7802Gain.NAU7802_GAIN_16,
    8: NAU7802Gain.NAU7802_GAIN_8,
    4: NAU7802Gain.NAU7802_GAIN_4,
    2: NAU7802Gain.NAU7802_GAIN_2,
    1: NAU7802Gain.NAU7802_GAIN_1,
}

NAU7802SPS = nau7802_ns.enum("NAU7802SPS")
SAMPLES_PER_SECOND = {
    320: NAU7802SPS.NAU7802_SPS_320,
    80: NAU7802SPS.NAU7802_SPS_80,
    40: NAU7802SPS.NAU7802_SPS_40,
    20: NAU7802SPS.NAU7802_SPS_20,
    10: NAU7802SPS.NAU7802_SPS_10,
}

NAU7802LDO = nau7802_ns.enum("NAU7802LDO")
LDO = {
    "2.4V": NAU7802LDO.NAU7802_LDO_2V4,
    "2.7V": NAU7802LDO.NAU7802_LDO_2V7,
    "3.0V": NAU7802LDO.NAU7802_LDO_3V0,
    "3.3V": NAU7802LDO.NAU7802_LDO_3V3,
    "3.6V": NAU7802LDO.NAU7802_LDO_3V6,
    "3.9V": NAU7802LDO.NAU7802_LDO_3V9,
    "4.2V": NAU7802LDO.NAU7802_LDO_4V2,
    "4.5V": NAU7802LDO.NAU7802_LDO_4V5,
    "EXTERNAL": NAU7802LDO.NAU7802_LDO_EXTERNAL,
    "EXT": NAU7802LDO.NAU7802_LDO_EXTERNAL,
}

CONFIG_SCHEMA = (
    sensor.sensor_schema(
        NAU7802Sensor,
        icon=ICON_SCALE,
        accuracy_decimals=0,
        state_class=STATE_CLASS_MEASUREMENT,
    )
    .extend(
        {
            cv.Optional(CONF_LDO_VOLTAGE, default="3.0V"): cv.enum(LDO, upper=True),
            cv.Optional(CONF_SAMPLES_PER_SECOND, default=10): cv.enum(
                SAMPLES_PER_SECOND, int=True
            ),
            cv.Optional(CONF_GAIN, default=128): cv.enum(GAINS, int=True),
            cv.Optional(CONF_OFFSET_CALIBRATION, default=0): cv.int_range(
                min=-8388608, max=8388607
            ),
            cv.Optional(CONF_GAIN_CALIBRATION, default=1.0): cv.float_range(
                min=0, max=511.9999998807907
            ),
        }
    )
    .extend(cv.polling_component_schema("60s"))
    .extend(i2c.i2c_device_schema(0x2A))
)


async def to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    await cg.register_component(var, config)
    await i2c.register_i2c_device(var, config)
    await sensor.register_sensor(var, config)

    cg.add(var.set_samples_per_second(config[CONF_SAMPLES_PER_SECOND]))
    cg.add(var.set_ldo_voltage(config[CONF_LDO_VOLTAGE]))
    cg.add(var.set_gain(config[CONF_GAIN]))
    cg.add(var.set_gain_calibration(config[CONF_GAIN_CALIBRATION]))
    cg.add(var.set_offset_calibration(config[CONF_OFFSET_CALIBRATION]))


NAU7802_CALIBRATE_SCHEMA = maybe_simple_id(
    {
        cv.GenerateID(CONF_ID): cv.use_id(NAU7802Sensor),
    }
)


@automation.register_action(
    "nau7802.calibrate_internal_offset",
    NAU7802CalbrateInternalOffsetAction,
    NAU7802_CALIBRATE_SCHEMA,
)
@automation.register_action(
    "nau7802.calibrate_external_offset",
    NAU7802CalbrateExternalOffsetAction,
    NAU7802_CALIBRATE_SCHEMA,
)
@automation.register_action(
    "nau7802.calibrate_gain",
    NAU7802CalbrateGainAction,
    NAU7802_CALIBRATE_SCHEMA,
)
async def nau7802_calibrate_to_code(config, action_id, template_arg, args):
    var = cg.new_Pvariable(action_id, template_arg)
    await cg.register_parented(var, config[CONF_ID])
    return var



import esphome.codegen as cg
from esphome.components import output
import esphome.config_validation as cv
from esphome.const import CONF_ID, CONF_LED

from ..display import CONF_TM1638_ID, TM1638Component, tm1638_ns

TM1638OutputLed = tm1638_ns.class_("TM1638OutputLed", output.BinaryOutput, cg.Component)


CONFIG_SCHEMA = output.BINARY_OUTPUT_SCHEMA.extend(
    {
        cv.GenerateID(): cv.declare_id(TM1638OutputLed),
        cv.GenerateID(CONF_TM1638_ID): cv.use_id(TM1638Component),
        cv.Required(CONF_LED): cv.int_range(min=0, max=7),
    }
).extend(cv.COMPONENT_SCHEMA)


async def to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    await output.register_output(var, config)
    await cg.register_component(var, config)
    cg.add(var.set_lednum(config[CONF_LED]))
    hub = await cg.get_variable(config[CONF_TM1638_ID])
    cg.add(var.set_tm1638(hub))



"""Constants for memory analysis symbol pattern matching."""

import re

# Pattern to extract ESPHome component namespaces dynamically
ESPHOME_COMPONENT_PATTERN = re.compile(r"esphome::([a-zA-Z0-9_]+)::")

# Section mapping for ELF file sections
# Maps standard section names to their various platform-specific variants
SECTION_MAPPING = {
    ".text": frozenset([".text", ".iram"]),
    ".rodata": frozenset([".rodata"]),
    ".data": frozenset([".data", ".dram"]),
    ".bss": frozenset([".bss"]),
}

# Section to ComponentMemory attribute mapping
# Maps section names to the attribute name in ComponentMemory dataclass
SECTION_TO_ATTR = {
    ".text": "text_size",
    ".rodata": "rodata_size",
    ".data": "data_size",
    ".bss": "bss_size",
}

# Component identification rules
# Symbol patterns: patterns found in raw symbol names
SYMBOL_PATTERNS = {
    "freertos": [
        "vTask",
        "xTask",
        "xQueue",
        "pvPort",
        "vPort",
        "uxTask",
        "pcTask",
        "prvTimerTask",
        "prvAddNewTaskToReadyList",
        "pxReadyTasksLists",
        "prvAddCurrentTaskToDelayedList",
        "xEventGroupWaitBits",
        "xRingbufferSendFromISR",
        "prvSendItemDoneNoSplit",
        "prvReceiveGeneric",
        "prvSendAcquireGeneric",
        "prvCopyItemAllowSplit",
        "xEventGroup",
        "xRingbuffer",
        "prvSend",
        "prvReceive",
        "prvCopy",
        "xPort",
        "ulTaskGenericNotifyTake",
        "prvIdleTask",
        "prvInitialiseNewTask",
        "prvIsYieldRequiredSMP",
        "prvGetItemByteBuf",
        "prvInitializeNewRingbuffer",
        "prvAcquireItemNoSplit",
        "prvNotifyQueueSetContainer",
        "ucStaticTimerQueueStorage",
        "eTaskGetState",
        "main_task",
        "do_system_init_fn",
        "xSemaphoreCreateGenericWithCaps",
        "vListInsert",
        "uxListRemove",
        "vRingbufferReturnItem",
        "vRingbufferReturnItemFromISR",
        "prvCheckItemFitsByteBuffer",
        "prvGetCurMaxSizeAllowSplit",
        "tick_hook",
        "sys_sem_new",
        "sys_arch_mbox_fetch",
        "sys_arch_sem_wait",
        "prvDeleteTCB",
        "vQueueDeleteWithCaps",
        "vRingbufferDeleteWithCaps",
        "vSemaphoreDeleteWithCaps",
        "prvCheckItemAvail",
        "prvCheckTaskCanBeScheduledSMP",
        "prvGetCurMaxSizeNoSplit",
        "prvResetNextTaskUnblockTime",
        "prvReturnItemByteBuf",
        "vApplicationStackOverflowHook",
        "vApplicationGetIdleTaskMemory",
        "sys_init",
        "sys_mbox_new",
        "sys_arch_mbox_tryfetch",
    ],
    "xtensa": ["xt_", "_xt_", "xPortEnterCriticalTimeout"],
    "heap": ["heap_", "multi_heap"],
    "spi_flash": ["spi_flash"],
    "rtc": ["rtc_", "rtcio_ll_"],
    "gpio_driver": ["gpio_", "pins"],
    "uart_driver": ["uart", "_uart", "UART"],
    "timer": ["timer_", "esp_timer"],
    "peripherals": ["periph_", "periman"],
    "network_stack": [
        "vj_compress",
        "raw_sendto",
        "raw_input",
        "etharp_",
        "icmp_input",
        "socket_ipv6",
        "ip_napt",
        "socket_ipv4_multicast",
        "socket_ipv6_multicast",
        "netconn_",
        "recv_raw",
        "accept_function",
        "netconn_recv_data",
        "netconn_accept",
        "netconn_write_vectors_partly",
        "netconn_drain",
        "raw_connect",
        "raw_bind",
        "icmp_send_response",
        "sockets",
        "icmp_dest_unreach",
        "inet_chksum_pseudo",
        "alloc_socket",
        "done_socket",
        "set_global_fd_sets",
        "inet_chksum_pbuf",
        "tryget_socket_unconn_locked",
        "tryget_socket_unconn",
        "cs_create_ctrl_sock",
        "netbuf_alloc",
        "tcp_",  # TCP protocol functions
        "udp_",  # UDP protocol functions
        "lwip_",  # LwIP stack functions
        "eagle_lwip",  # ESP-specific LwIP functions
        "new_linkoutput",  # Link output function
        "acd_",  # Address Conflict Detection (ACD)
        "eth_",  # Ethernet functions
        "mac_enable_bb",  # MAC baseband enable
        "reassemble_and_dispatch",  # Packet reassembly
    ],
    # dhcp must come before libc to avoid "dhcp_select" matching "select" pattern
    "dhcp": ["dhcp", "handle_dhcp"],
    "ipv6_stack": ["nd6_", "ip6_", "mld6_", "icmp6_", "icmp6_input"],
    # Order matters! More specific categories must come before general ones.
    # mdns must come before bluetooth to avoid "_mdns_disable_pcb" matching "ble_" pattern
    "mdns_lib": ["mdns"],
    # memory_mgmt must come before wifi_stack to catch mmu_hal_* symbols
    "memory_mgmt": [
        "mem_",
        "memory_",
        "tlsf_",
        "memp_",
        "pbuf_",
        "pbuf_alloc",
        "pbuf_copy_partial_pbuf",
        "esp_mmu_map",
        "mmu_hal_",
        "s_do_mapping",  # Memory mapping function, not WiFi
        "hash_map_",  # Hash map data structure
        "umm_assimilate",  # UMM malloc assimilation
    ],
    # Bluetooth categories must come BEFORE wifi_stack to avoid misclassification
    # Many BLE symbols contain patterns like "ble_" that would otherwise match wifi patterns
    "bluetooth_rom": ["r_ble", "r_lld", "r_llc", "r_llm"],
    "bluedroid_bt": [
        "bluedroid",
        "btc_",
        "bta_",
        "btm_",
        "btu_",
        "BTM_",
        "GATT",
        "L2CA_",
        "smp_",
        "gatts_",
        "attp_",
        "l2cu_",
        "l2cb",
        "smp_cb",
        "BTA_GATTC_",
        "SMP_",
        "BTU_",
        "BTA_Dm",
        "GAP_Ble",
        "BT_tx_if",
        "host_recv_pkt_cb",
        "saved_local_oob_data",
        "string_to_bdaddr",
        "string_is_bdaddr",
        "CalConnectParamTimeout",
        "transmit_fragment",
        "transmit_data",
        "event_command_ready",
        "read_command_complete_header",
        "parse_read_local_extended_features_response",
        "parse_read_local_version_info_response",
        "should_request_high",
        "btdm_wakeup_request",
        "BTA_SetAttributeValue",
        "BTA_EnableBluetooth",
        "transmit_command_futured",
        "transmit_command",
        "get_waiting_command",
        "make_command",
        "transmit_downward",
        "host_recv_adv_packet",
        "copy_extra_byte_in_db",
        "parse_read_local_supported_commands_response",
    ],
    "bluetooth": [
        "bt_",
        "_ble_",  # More specific than "ble_" to avoid matching "able_", "enable_", "disable_"
        "l2c_",
        "l2ble_",  # L2CAP for BLE
        "gatt_",
        "gap_",
        "hci_",
        "btsnd_hcic_",  # Bluetooth HCI command send functions
        "BT_init",
        "BT_tx_",  # Bluetooth transmit functions
        "esp_ble_",  # Catch esp_ble_* functions
    ],
    "bluetooth_ll": [
        "llm_",  # Link layer manager
        "llc_",  # Link layer control
        "lld_",  # Link layer driver
        "ld_acl_",  # Link layer ACL (Asynchronous Connection-Oriented)
        "llcp_",  # Link layer control protocol
        "lmp_",  # Link manager protocol
    ],
    "wifi_bt_coex": ["coex"],
    "wifi_stack": [
        "ieee80211",
        "hostap",
        "sta_",
        "wifi_ap_",  # More specific than "ap_" to avoid matching "cap_", "map_"
        "wifi_scan_",  # More specific than "scan_" to avoid matching "_scan_" in other contexts
        "wifi_",
        "wpa_",
        "wps_",
        "esp_wifi",
        "cnx_",
        "wpa3_",
        "sae_",
        "wDev_",
        "ic_mac_",  # More specific than "mac_" to avoid matching emac_
        "esf_buf",
        "gWpaSm",
        "sm_WPA",
        "eapol_",
        "owe_",
        "wifiLowLevelInit",
        # Removed "s_do_mapping" - this is memory management, not WiFi
        "gScanStruct",
        "ppSearchTxframe",
        "ppMapWaitTxq",
        "ppFillAMPDUBar",
        "ppCheckTxConnTrafficIdle",
        "ppCalTkipMic",
        "phy_force_wifi",
        "phy_unforce_wifi",
        "write_wifi_chan",
        "wifi_track_pll",
    ],
    "crypto_math": [
        "ecp_",
        "bignum_",
        "mpi_",
        "sswu",
        "modp",
        "dragonfly_",
        "gcm_mult",
        "__multiply",
        "quorem",
        "__mdiff",
        "__lshift",
        "__mprec_tens",
        "ECC_",
        "multiprecision_",
        "mix_sub_columns",
        "sbox",
        "gfm2_sbox",
        "gfm3_sbox",
        "curve_p256",
        "curve",
        "p_256_init_curve",
        "shift_sub_rows",
        "rshift",
        "rijndaelEncrypt",  # AES Rijndael encryption
    ],
    # System and Arduino core functions must come before libc
    "esp_system": [
        "system_",  # ESP system functions
        "postmortem_",  # Postmortem reporting
    ],
    "arduino_core": [
        "pinMode",
        "resetPins",
        "millis",
        "micros",
        "delay(",  # More specific - Arduino delay function with parenthesis
        "delayMicroseconds",
        "digitalWrite",
        "digitalRead",
    ],
    "sntp": ["sntp_", "sntp_recv"],
    "scheduler": [
        "run_scheduled_",
        "compute_scheduled_",
        "event_TaskQueue",
    ],
    "hw_crypto": ["esp_aes", "esp_sha", "esp_rsa", "esp_bignum", "esp_mpi"],
    "libc": [
        "printf",
        "scanf",
        "malloc",
        "_free",  # More specific than "free" to match _free, __free_r, etc. but not arbitrary "free" substring
        "umm_free",  # UMM malloc free function
        "memcpy",
        "memset",
        "strcpy",
        "strlen",
        "_dtoa",
        "_fopen",
        "__sfvwrite_r",
        "qsort",
        "__sf",
        "__sflush_r",
        "__srefill_r",
        "_impure_data",
        "_reclaim_reent",
        "_open_r",
        "strncpy",
        "_strtod_l",
        "__gethex",
        "__hexnan",
        "_setenv_r",
        "_tzset_unlocked_r",
        "__tzcalc_limits",
        "_select",  # More specific than "select" to avoid matching "dhcp_select", etc.
        "scalbnf",
        "strtof",
        "strtof_l",
        "__d2b",
        "__b2d",
        "__s2b",
        "_Balloc",
        "__multadd",
        "__lo0bits",
        "__atexit0",
        "__smakebuf_r",
        "__swhatbuf_r",
        "_sungetc_r",
        "_close_r",
        "_link_r",
        "_unsetenv_r",
        "_rename_r",
        "__month_lengths",
        "tzinfo",
        "__ratio",
        "__hi0bits",
        "__ulp",
        "__any_on",
        "__copybits",
        "L_shift",
        "_fcntl_r",
        "_lseek_r",
        "_read_r",
        "_write_r",
        "_unlink_r",
        "_fstat_r",
        "access",
        "fsync",
        "tcsetattr",
        "tcgetattr",
        "tcflush",
        "tcdrain",
        "__ssrefill_r",
        "_stat_r",
        "__hexdig_fun",
        "__mcmp",
        "_fwalk_sglue",
        "__fpclassifyf",
        "_setlocale_r",
        "_mbrtowc_r",
        "fcntl",
        "__match",
        "_lock_close",
        "__c$",
        "__func__$",
        "__FUNCTION__$",
        "DAYS_IN_MONTH",
        "_DAYS_BEFORE_MONTH",
        "CSWTCH$",
        "dst$",
        "sulp",
        "_strtol_l",  # String to long with locale
        "__cvt",  # Convert
        "__utoa",  # Unsigned to ASCII
        "__global_locale",  # Global locale
        "_ctype_",  # Character type
        "impure_data",  # Impure data
    ],
    "string_ops": [
        "strcmp",
        "strncmp",
        "strchr",
        "strstr",
        "strtok",
        "strdup",
        "strncasecmp_P",  # String compare (case insensitive, from program memory)
        "strnlen_P",  # String length (from program memory)
        "strncat_P",  # String concatenate (from program memory)
    ],
    "memory_alloc": ["malloc", "calloc", "realloc", "free", "_sbrk"],
    "file_io": [
        "fread",
        "fwrite",
        "fopen",
        "fclose",
        "fseek",
        "ftell",
        "fflush",
        "s_fd_table",
    ],
    "string_formatting": [
        "snprintf",
        "vsnprintf",
        "sprintf",
        "vsprintf",
        "sscanf",
        "vsscanf",
    ],
    "cpp_anonymous": ["_GLOBAL__N_", "n$"],
    # Plain C patterns only - C++ symbols will be categorized via DEMANGLED_PATTERNS
    "nvs": ["nvs_"],  # Plain C NVS functions
    "ota": ["ota_", "OTA", "esp_ota", "app_desc"],
    # cpp_runtime: Removed _ZN, _ZL to let DEMANGLED_PATTERNS categorize C++ symbols properly
    # Only keep patterns that are truly runtime-specific and not categorizable by namespace
    "cpp_runtime": ["__cxx", "_ZSt", "__gxx_personality", "_Z16"],
    "exception_handling": [
        "__cxa_",
        "_Unwind_",
        "__gcc_personality",
        "uw_frame_state",
        "search_object",  # Search for exception handling object
        "get_cie_encoding",  # Get CIE encoding
        "add_fdes",  # Add frame description entries
        "fde_unencoded_compare",  # Compare FDEs
        "fde_mixed_encoding_compare",  # Compare mixed encoding FDEs
        "frame_downheap",  # Frame heap operations
        "frame_heapsort",  # Frame heap sorting
    ],
    "static_init": ["_GLOBAL__sub_I_"],
    "phy_radio": [
        "phy_",
        "rf_",
        "chip_",
        "register_chipv7",
        "pbus_",
        "bb_",
        "fe_",
        "rfcal_",
        "ram_rfcal",
        "tx_pwctrl",
        "rx_chan",
        "set_rx_gain",
        "set_chan",
        "agc_reg",
        "ram_txiq",
        "ram_txdc",
        "ram_gen_rx_gain",
        "rx_11b_opt",
        "set_rx_sense",
        "set_rx_gain_cal",
        "set_chan_dig_gain",
        "tx_pwctrl_init_cal",
        "rfcal_txiq",
        "set_tx_gain_table",
        "correct_rfpll_offset",
        "pll_correct_dcap",
        "txiq_cal_init",
        "pwdet_sar",
        "pwdet_sar2_init",
        "ram_iq_est_enable",
        "ram_rfpll_set_freq",
        "ant_wifirx_cfg",
        "ant_btrx_cfg",
        "force_txrxoff",
        "force_txrx_off",
        "tx_paon_set",
        "opt_11b_resart",
        "rfpll_1p2_opt",
        "ram_dc_iq_est",
        "ram_start_tx_tone",
        "ram_en_pwdet",
        "ram_cbw2040_cfg",
        "rxdc_est_min",
        "i2cmst_reg_init",
        "temprature_sens_read",
        "ram_restart_cal",
        "ram_write_gain_mem",
        "ram_wait_rfpll_cal_end",
        "txcal_debuge_mode",
        "ant_wifitx_cfg",
        "reg_init_begin",
        "tx_cap_init",  # TX capacitance init
        "ram_set_txcap",  # RAM TX capacitance setting
        "tx_atten_",  # TX attenuation
        "txiq_",  # TX I/Q calibration
        "ram_cal_",  # RAM calibration
        "ram_rxiq_",  # RAM RX I/Q
        "readvdd33",  # Read VDD33
        "test_tout",  # Test timeout
        "tsen_meas",  # Temperature sensor measurement
        "bbpll_cal",  # Baseband PLL calibration
        "set_cal_",  # Set calibration
        "set_rfanagain_",  # Set RF analog gain
        "set_txdc_",  # Set TX DC
        "get_vdd33_",  # Get VDD33
        "gen_rx_gain_table",  # Generate RX gain table
        "ram_ana_inf_gating_en",  # RAM analog interface gating enable
        "tx_cont_en",  # TX continuous enable
        "tx_delay_cfg",  # TX delay configuration
        "tx_gain_table_set",  # TX gain table set
        "check_and_reset_hw_deadlock",  # Hardware deadlock check
        "s_config",  # System/hardware config
        "chan14_mic_cfg",  # Channel 14 MIC config
    ],
    "wifi_phy_pp": [
        "pp_",
        "ppT",
        "ppR",
        "ppP",
        "ppInstall",
        "ppCalTxAMPDULength",
        "ppCheckTx",  # Packet processor TX check
        "ppCal",  # Packet processor calibration
        "HdlAllBuffedEb",  # Handle buffered EB
    ],
    "wifi_lmac": ["lmac"],
    "wifi_device": [
        "wdev",
        "wDev_",
        "ic_set_sta",  # Set station mode
        "ic_set_vif",  # Set virtual interface
    ],
    "power_mgmt": [
        "pm_",
        "sleep",
        "rtc_sleep",
        "light_sleep",
        "deep_sleep",
        "power_down",
        "g_pm",
        "pmc",  # Power Management Controller
    ],
    "hal_layer": ["hal_"],
    "clock_mgmt": [
        "clk_",
        "clock_",
        "rtc_clk",
        "apb_",
        "cpu_freq",
        "setCpuFrequencyMhz",
    ],
    "cache_mgmt": ["cache"],
    "flash_ops": ["flash", "image_load"],
    "interrupt_handlers": [
        "isr",
        "interrupt",
        "intr_",
        "exc_",
        "exception",
        "port_IntStack",
    ],
    "wrapper_functions": ["_wrapper"],
    "error_handling": ["panic", "abort", "assert", "error_", "fault"],
    "authentication": ["auth"],
    "ppp_protocol": ["ppp", "ipcp_", "lcp_", "chap_", "LcpEchoCheck"],
    "ethernet_phy": [
        "emac_",
        "eth_phy_",
        "phy_tlk110",
        "phy_lan87",
        "phy_ip101",
        "phy_rtl",
        "phy_dp83",
        "phy_ksz",
        "lan87xx_",
        "rtl8201_",
        "ip101_",
        "ksz80xx_",
        "jl1101_",
        "dp83848_",
        "eth_on_state_changed",
    ],
    "threading": ["pthread_", "thread_", "_task_"],
    "pthread": ["pthread"],
    "synchronization": ["mutex", "semaphore", "spinlock", "portMUX"],
    "math_lib": [
        "sin",
        "cos",
        "tan",
        "sqrt",
        "pow",
        "exp",
        "log",
        "atan",
        "asin",
        "acos",
        "floor",
        "ceil",
        "fabs",
        "round",
    ],
    "random": ["rand", "random", "rng_", "prng"],
    "time_lib": [
        "time",
        "clock",
        "gettimeofday",
        "settimeofday",
        "localtime",
        "gmtime",
        "mktime",
        "strftime",
    ],
    "console_io": ["console_", "uart_tx", "uart_rx", "puts", "putchar", "getchar"],
    "rom_functions": ["r_", "rom_"],
    "compiler_runtime": [
        "__divdi3",
        "__udivdi3",
        "__moddi3",
        "__muldi3",
        "__ashldi3",
        "__ashrdi3",
        "__lshrdi3",
        "__cmpdi2",
        "__fixdfdi",
        "__floatdidf",
    ],
    "libgcc": ["libgcc", "_divdi3", "_udivdi3"],
    "boot_startup": ["boot", "start_cpu", "call_start", "startup", "bootloader"],
    "bootloader": ["bootloader_", "esp_bootloader"],
    "app_framework": ["app_", "initArduino", "setup", "loop", "Update"],
    "weak_symbols": ["__weak_"],
    "compiler_builtins": ["__builtin_"],
    "vfs": ["vfs_", "VFS"],
    "esp32_sdk": ["esp32_", "esp32c", "esp32s"],
    "usb": ["usb_", "USB", "cdc_", "CDC"],
    "i2c_driver": ["i2c_", "I2C"],
    "i2s_driver": ["i2s_", "I2S"],
    "spi_driver": ["spi_", "SPI"],
    "adc_driver": ["adc_", "ADC"],
    "dac_driver": ["dac_", "DAC"],
    "touch_driver": ["touch_", "TOUCH"],
    "pwm_driver": ["pwm_", "PWM", "ledc_", "LEDC"],
    "rmt_driver": ["rmt_", "RMT"],
    "pcnt_driver": ["pcnt_", "PCNT"],
    "can_driver": ["can_", "CAN", "twai_", "TWAI"],
    "sdmmc_driver": ["sdmmc_", "SDMMC", "sdcard", "sd_card"],
    "temp_sensor": ["temp_sensor", "tsens_"],
    "watchdog": ["wdt_", "WDT", "watchdog"],
    "brownout": ["brownout", "bod_"],
    "ulp": ["ulp_", "ULP"],
    "psram": ["psram", "PSRAM", "spiram", "SPIRAM"],
    "efuse": ["efuse", "EFUSE"],
    "partition": ["partition", "esp_partition"],
    "esp_event": ["esp_event", "event_loop", "event_callback"],
    "esp_console": ["esp_console", "console_"],
    "chip_specific": ["chip_", "esp_chip"],
    "esp_system_utils": ["esp_system", "esp_hw", "esp_clk", "esp_sleep"],
    "ipc": ["esp_ipc", "ipc_"],
    "wifi_config": [
        "g_cnxMgr",
        "gChmCxt",
        "g_ic",
        "TxRxCxt",
        "s_dp",
        "s_ni",
        "s_reg_dump",
        "packet$",
        "d_mult_table",
        "K",
        "fcstab",
    ],
    "smartconfig": ["sc_ack_send"],
    "rc_calibration": ["rc_cal", "rcUpdate"],
    "noise_floor": ["noise_check"],
    "rf_calibration": [
        "set_rx_sense",
        "set_rx_gain_cal",
        "set_chan_dig_gain",
        "tx_pwctrl_init_cal",
        "rfcal_txiq",
        "set_tx_gain_table",
        "correct_rfpll_offset",
        "pll_correct_dcap",
        "txiq_cal_init",
        "pwdet_sar",
        "rx_11b_opt",
    ],
    "wifi_crypto": [
        "pk_use_ecparams",
        "process_segments",
        "ccmp_",
        "rc4_",
        "aria_",
        "mgf_mask",
        "dh_group",
        "ccmp_aad_nonce",
        "ccmp_encrypt",
        "rc4_skip",
        "aria_sb1",
        "aria_sb2",
        "aria_is1",
        "aria_is2",
        "aria_sl",
        "aria_a",
    ],
    "radio_control": ["fsm_input", "fsm_sconfreq"],
    "pbuf": [
        "pbuf_",
    ],
    "event_group": ["xEventGroup"],
    "ringbuffer": ["xRingbuffer", "prvSend", "prvReceive", "prvCopy"],
    "provisioning": ["prov_", "prov_stop_and_notify"],
    "scan": ["gScanStruct"],
    "port": ["xPort"],
    "elf_loader": [
        "elf_add",
        "elf_add_note",
        "elf_add_segment",
        "process_image",
        "read_encoded",
        "read_encoded_value",
        "read_encoded_value_with_base",
        "process_image_header",
    ],
    "socket_api": [
        "sockets",
        "netconn_",
        "accept_function",
        "recv_raw",
        "socket_ipv4_multicast",
        "socket_ipv6_multicast",
    ],
    "igmp": ["igmp_", "igmp_send", "igmp_input"],
    "icmp6": ["icmp6_"],
    "arp": ["arp_table"],
    "ampdu": [
        "ampdu_",
        "rcAmpdu",
        "trc_onAmpduOp",
        "rcAmpduLowerRate",
        "ampdu_dispatch_upto",
    ],
    "ieee802_11": ["ieee802_11_", "ieee802_11_parse_elems"],
    "rate_control": [
        "rssi_margin",
        "rcGetSched",
        "get_rate_fcc_index",
        "rcGetRate",  # Get rate
        "rc_get_",  # Rate control getters
        "rc_set_",  # Rate control setters
        "rc_enable_",  # Rate control enable functions
    ],
    "nan": ["nan_dp_", "nan_dp_post_tx", "nan_dp_delete_peer"],
    "channel_mgmt": ["chm_init", "chm_set_current_channel"],
    "trace": ["trc_init", "trc_onAmpduOp"],
    "country_code": ["country_info", "country_info_24ghz"],
    "multicore": ["do_multicore_settings"],
    "Update_lib": ["Update"],
    "stdio": [
        "__sf",
        "__sflush_r",
        "__srefill_r",
        "_impure_data",
        "_reclaim_reent",
        "_open_r",
    ],
    "strncpy_ops": ["strncpy"],
    "math_internal": ["__mdiff", "__lshift", "__mprec_tens", "quorem"],
    "character_class": ["__chclass"],
    "camellia": ["camellia_", "camellia_feistel"],
    "crypto_tables": ["FSb", "FSb2", "FSb3", "FSb4"],
    "event_buffer": ["g_eb_list_desc", "eb_space"],
    "base_node": ["base_node_", "base_node_add_handler"],
    "file_descriptor": ["s_fd_table"],
    "tx_delay": ["tx_delay_cfg"],
    "deinit": ["deinit_functions"],
    "lcp_echo": ["LcpEchoCheck"],
    "raw_api": ["raw_bind", "raw_connect"],
    "checksum": ["process_checksum"],
    "entry_management": ["add_entry"],
    "esp_ota": ["esp_ota", "ota_", "read_otadata"],
    "http_server": [
        "httpd_",
        "parse_url_char",
        "cb_headers_complete",
        "delete_entry",
        "validate_structure",
        "config_save",
        "config_new",
        "verify_url",
        "cb_url",
    ],
    "misc_system": [
        "alarm_cbs",
        "start_up",
        "tokens",
        "unhex",
        "osi_funcs_ro",
        "enum_function",
        "fragment_and_dispatch",
        "alarm_set",
        "osi_alarm_new",
        "config_set_string",
        "config_update_newest_section",
        "config_remove_key",
        "method_strings",
        "interop_match",
        "interop_database",
        "__state_table",
        "__action_table",
        "s_stub_table",
        "s_context",
        "s_mmu_ctx",
        "s_get_bus_mask",
        "hli_queue_put",
        "list_remove",
        "list_delete",
        "lock_acquire_generic",
        "is_vect_desc_usable",
        "io_mode_str",
        "__c$20233",
        "interface",
        "read_id_core",
        "subscribe_idle",
        "unsubscribe_idle",
        "s_clkout_handle",
        "lock_release_generic",
        "config_set_int",
        "config_get_int",
        "config_get_string",
        "config_has_key",
        "config_remove_section",
        "osi_alarm_init",
        "osi_alarm_deinit",
        "fixed_queue_enqueue",
        "fixed_queue_dequeue",
        "fixed_queue_new",
        "fixed_pkt_queue_enqueue",
        "fixed_pkt_queue_new",
        "list_append",
        "list_prepend",
        "list_insert_after",
        "list_contains",
        "list_get_node",
        "hash_function_blob",
        "cb_no_body",
        "cb_on_body",
        "profile_tab",
        "get_arg",
        "trim",
        "buf$",
        "process_appended_hash_and_sig$constprop$0",
        "uuidType",
        "allocate_svc_db_buf",
        "_hostname_is_ours",
        "s_hli_handlers",
        "tick_cb",
        "idle_cb",
        "input",
        "entry_find",
        "section_find",
        "find_bucket_entry_",
        "config_has_section",
        "hli_queue_create",
        "hli_queue_get",
        "hli_c_handler",
        "future_ready",
        "future_await",
        "future_new",
        "pkt_queue_enqueue",
        "pkt_queue_dequeue",
        "pkt_queue_cleanup",
        "pkt_queue_create",
        "pkt_queue_destroy",
        "fixed_pkt_queue_dequeue",
        "osi_alarm_cancel",
        "osi_alarm_is_active",
        "osi_sem_take",
        "osi_event_create",
        "osi_event_bind",
        "alarm_cb_handler",
        "list_foreach",
        "list_back",
        "list_front",
        "list_clear",
        "fixed_queue_try_peek_first",
        "translate_path",
        "get_idx",
        "find_key",
        "init",
        "end",
        "start",
        "set_read_value",
        "copy_address_list",
        "copy_and_key",
        "sdk_cfg_opts",
        "leftshift_onebit",
        "config_section_end",
        "config_section_begin",
        "find_entry_and_check_all_reset",
        "image_validate",
        "xPendingReadyList",
        "vListInitialise",
        "lock_init_generic",
        "ant_bttx_cfg",
        "ant_dft_cfg",
        "cs_send_to_ctrl_sock",
        "config_llc_util_funcs_reset",
        "make_set_adv_report_flow_control",
        "make_set_event_mask",
        "raw_new",
        "raw_remove",
        "BTE_InitStack",
        "parse_read_local_supported_features_response",
        "__math_invalidf",
        "tinytens",
        "__mprec_tinytens",
        "__mprec_bigtens",
        "vRingbufferDelete",
        "vRingbufferDeleteWithCaps",
        "vRingbufferReturnItem",
        "vRingbufferReturnItemFromISR",
        "get_acl_data_size_ble",
        "get_features_ble",
        "get_features_classic",
        "get_acl_packet_size_ble",
        "get_acl_packet_size_classic",
        "supports_extended_inquiry_response",
        "supports_rssi_with_inquiry_results",
        "supports_interlaced_inquiry_scan",
        "supports_reading_remote_extended_features",
    ],
}

# Demangled patterns: patterns found in demangled C++ names
DEMANGLED_PATTERNS = {
    "gpio_driver": ["GPIO"],
    "uart_driver": ["UART"],
    # mdns_lib must come before network_stack to avoid "udp" matching "_udpReadBuffer" in MDNSResponder
    "mdns_lib": [
        "MDNSResponder",
        "MDNSImplementation",
        "MDNS",
    ],
    "network_stack": [
        "lwip",
        "tcp",
        "udp",
        "ip4",
        "ip6",
        "dhcp",
        "dns",
        "netif",
        "ethernet",
        "ppp",
        "slip",
        "UdpContext",  # UDP context class
        "DhcpServer",  # DHCP server class
    ],
    "arduino_core": [
        "String::",  # Arduino String class
        "Print::",  # Arduino Print class
        "HardwareSerial::",  # Serial class
        "IPAddress::",  # IP address class
        "EspClass::",  # ESP class
        "experimental::_SPI",  # Experimental SPI
    ],
    "ota": [
        "UpdaterClass",
        "Updater::",
    ],
    "wifi": [
        "ESP8266WiFi",
        "WiFi::",
    ],
    "wifi_stack": ["NetworkInterface"],
    "nimble_bt": [
        "nimble",
        "NimBLE",
        "ble_hs",
        "ble_gap",
        "ble_gatt",
        "ble_att",
        "ble_l2cap",
        "ble_sm",
    ],
    "crypto": ["mbedtls", "crypto", "sha", "aes", "rsa", "ecc", "tls", "ssl"],
    "cpp_stdlib": ["std::", "__gnu_cxx::", "__cxxabiv"],
    "static_init": ["__static_initialization"],
    "rtti": ["__type_info", "__class_type_info"],
    "web_server_lib": ["AsyncWebServer", "AsyncWebHandler", "WebServer"],
    "async_tcp": ["AsyncClient", "AsyncServer"],
    "json_lib": [
        "ArduinoJson",
        "JsonDocument",
        "JsonArray",
        "JsonObject",
        "deserialize",
        "serialize",
    ],
    "http_lib": ["HTTP", "http_", "Request", "Response", "Uri", "WebSocket"],
    "logging": ["log", "Log", "print", "Print", "diag_"],
    "authentication": ["checkDigestAuthentication"],
    "libgcc": ["libgcc"],
    "esp_system": ["esp_", "ESP"],
    "arduino": ["arduino"],
    "nvs": ["nvs_", "_ZTVN3nvs", "nvs::"],
    "filesystem": ["spiffs", "vfs"],
    "libc": ["newlib"],
}

# Patterns for categorizing ESPHome core symbols into subcategories
CORE_SUBCATEGORY_PATTERNS = {
    "Component Framework": ["Component"],
    "Application Core": ["Application"],
    "Scheduler": ["Scheduler"],
    "Component Iterator": ["ComponentIterator"],
    "Helper Functions": ["Helpers", "helpers"],
    "Preferences/Storage": ["Preferences", "ESPPreferences"],
    "I/O Utilities": ["HighFrequencyLoopRequester"],
    "String Utilities": ["str_"],
    "Bit Utilities": ["reverse_bits"],
    "Data Conversion": ["convert_"],
    "Network Utilities": ["network", "IPAddress"],
    "API Protocol": ["api::"],
    "WiFi Manager": ["wifi::"],
    "MQTT Client": ["mqtt::"],
    "Logger": ["logger::"],
    "OTA Updates": ["ota::"],
    "Web Server": ["web_server::"],
    "Time Management": ["time::"],
    "Sensor Framework": ["sensor::"],
    "Binary Sensor": ["binary_sensor::"],
    "Switch Framework": ["switch_::"],
    "Light Framework": ["light::"],
    "Climate Framework": ["climate::"],
    "Cover Framework": ["cover::"],
}



import esphome.codegen as cg
from esphome.components.packet_transport import (
    PacketTransport,
    new_packet_transport,
    transport_schema,
)
import esphome.config_validation as cv
from esphome.cpp_types import PollingComponent

from .. import CONF_SX126X_ID, SX126x, SX126xListener, sx126x_ns

SX126xTransport = sx126x_ns.class_(
    "SX126xTransport", PacketTransport, PollingComponent, SX126xListener
)

CONFIG_SCHEMA = transport_schema(SX126xTransport).extend(
    {
        cv.GenerateID(CONF_SX126X_ID): cv.use_id(SX126x),
    }
)


async def to_code(config):
    var, _ = await new_packet_transport(config)
    sx126x = await cg.get_variable(config[CONF_SX126X_ID])
    cg.add(var.set_parent(sx126x))



from esphome import automation
from esphome.automation import maybe_simple_id
import esphome.codegen as cg
from esphome.components import i2c, sensor
import esphome.config_validation as cv
from esphome.const import (
    CONF_BATTERY_LEVEL,
    CONF_BATTERY_VOLTAGE,
    CONF_ID,
    DEVICE_CLASS_BATTERY,
    DEVICE_CLASS_VOLTAGE,
    ENTITY_CATEGORY_DIAGNOSTIC,
    STATE_CLASS_MEASUREMENT,
    UNIT_PERCENT,
    UNIT_VOLT,
)

DEPENDENCIES = ["i2c"]

max17043_ns = cg.esphome_ns.namespace("max17043")
MAX17043Component = max17043_ns.class_(
    "MAX17043Component", cg.PollingComponent, i2c.I2CDevice
)

# Actions
SleepAction = max17043_ns.class_("SleepAction", automation.Action)

CONFIG_SCHEMA = (
    cv.Schema(
        {
            cv.GenerateID(): cv.declare_id(MAX17043Component),
            cv.Optional(CONF_BATTERY_VOLTAGE): sensor.sensor_schema(
                unit_of_measurement=UNIT_VOLT,
                accuracy_decimals=3,
                device_class=DEVICE_CLASS_VOLTAGE,
                state_class=STATE_CLASS_MEASUREMENT,
                entity_category=ENTITY_CATEGORY_DIAGNOSTIC,
            ),
            cv.Optional(CONF_BATTERY_LEVEL): sensor.sensor_schema(
                unit_of_measurement=UNIT_PERCENT,
                accuracy_decimals=3,
                device_class=DEVICE_CLASS_BATTERY,
                state_class=STATE_CLASS_MEASUREMENT,
                entity_category=ENTITY_CATEGORY_DIAGNOSTIC,
            ),
        }
    )
    .extend(cv.polling_component_schema("60s"))
    .extend(i2c.i2c_device_schema(0x36))
)


async def to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    await cg.register_component(var, config)
    await i2c.register_i2c_device(var, config)

    if voltage_config := config.get(CONF_BATTERY_VOLTAGE):
        sens = await sensor.new_sensor(voltage_config)
        cg.add(var.set_voltage_sensor(sens))

    if CONF_BATTERY_LEVEL in config:
        sens = await sensor.new_sensor(config[CONF_BATTERY_LEVEL])
        cg.add(var.set_battery_remaining_sensor(sens))


MAX17043_ACTION_SCHEMA = maybe_simple_id(
    {
        cv.Required(CONF_ID): cv.use_id(MAX17043Component),
    }
)


@automation.register_action("max17043.sleep_mode", SleepAction, MAX17043_ACTION_SCHEMA)
async def max17043_sleep_mode_to_code(config, action_id, template_arg, args):
    paren = await cg.get_variable(config[CONF_ID])
    return cg.new_Pvariable(action_id, template_arg, paren)



from esphome import automation
import esphome.codegen as cg
from esphome.components import microphone, sensor
import esphome.config_validation as cv
from esphome.const import (
    CONF_ID,
    CONF_MEASUREMENT_DURATION,
    CONF_MICROPHONE,
    DEVICE_CLASS_SOUND_PRESSURE,
    PLATFORM_ESP32,
    STATE_CLASS_MEASUREMENT,
    UNIT_DECIBEL,
)

AUTO_LOAD = ["audio"]
CODEOWNERS = ["@kahrendt"]
DEPENDENCIES = ["microphone"]


CONF_PASSIVE = "passive"
CONF_PEAK = "peak"
CONF_RMS = "rms"

sound_level_ns = cg.esphome_ns.namespace("sound_level")
SoundLevelComponent = sound_level_ns.class_("SoundLevelComponent", cg.Component)

StartAction = sound_level_ns.class_("StartAction", automation.Action)
StopAction = sound_level_ns.class_("StopAction", automation.Action)

CONFIG_SCHEMA = cv.All(
    cv.Schema(
        {
            cv.GenerateID(): cv.declare_id(SoundLevelComponent),
            cv.Optional(CONF_MEASUREMENT_DURATION, default="1000ms"): cv.All(
                cv.positive_time_period_milliseconds,
                cv.Range(
                    min=cv.TimePeriod(milliseconds=50),
                    max=cv.TimePeriod(seconds=60),
                ),
            ),
            cv.Optional(
                CONF_MICROPHONE, default={}
            ): microphone.microphone_source_schema(
                min_bits_per_sample=16,
                max_bits_per_sample=16,
            ),
            cv.Required(CONF_PASSIVE): cv.boolean,
            cv.Optional(CONF_PEAK): sensor.sensor_schema(
                unit_of_measurement=UNIT_DECIBEL,
                accuracy_decimals=1,
                device_class=DEVICE_CLASS_SOUND_PRESSURE,
                state_class=STATE_CLASS_MEASUREMENT,
            ),
            cv.Optional(CONF_RMS): sensor.sensor_schema(
                unit_of_measurement=UNIT_DECIBEL,
                accuracy_decimals=1,
                device_class=DEVICE_CLASS_SOUND_PRESSURE,
                state_class=STATE_CLASS_MEASUREMENT,
            ),
        }
    ).extend(cv.COMPONENT_SCHEMA),
    cv.only_on([PLATFORM_ESP32]),
)


async def to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    await cg.register_component(var, config)

    mic_source = await microphone.microphone_source_to_code(
        config[CONF_MICROPHONE], passive=config[CONF_PASSIVE]
    )
    cg.add(var.set_microphone_source(mic_source))

    cg.add(var.set_measurement_duration(config[CONF_MEASUREMENT_DURATION]))

    if peak_config := config.get(CONF_PEAK):
        sens = await sensor.new_sensor(peak_config)
        cg.add(var.set_peak_sensor(sens))
    if rms_config := config.get(CONF_RMS):
        sens = await sensor.new_sensor(rms_config)
        cg.add(var.set_rms_sensor(sens))


SOUND_LEVEL_ACTION_SCHEMA = automation.maybe_simple_id(
    {
        cv.GenerateID(): cv.use_id(SoundLevelComponent),
    }
)


@automation.register_action("sound_level.start", StartAction, SOUND_LEVEL_ACTION_SCHEMA)
@automation.register_action("sound_level.stop", StopAction, SOUND_LEVEL_ACTION_SCHEMA)
async def sound_level_action_to_code(config, action_id, template_arg, args):
    var = cg.new_Pvariable(action_id, template_arg)
    await cg.register_parented(var, config[CONF_ID])
    return var



import esphome.codegen as cg
from esphome.components import sensor
import esphome.config_validation as cv
from esphome.const import (
    CONF_DISTANCE,
    DEVICE_CLASS_DISTANCE,
    ICON_COUNTER,
    ICON_HEART_PULSE,
    ICON_PULSE,
    ICON_SIGNAL,
    STATE_CLASS_MEASUREMENT,
    UNIT_BEATS_PER_MINUTE,
    UNIT_CENTIMETER,
)

from . import CONF_MR60BHA2_ID, MR60BHA2Component

DEPENDENCIES = ["seeed_mr60bha2"]

CONF_BREATH_RATE = "breath_rate"
CONF_HEART_RATE = "heart_rate"
CONF_NUM_TARGETS = "num_targets"

CONFIG_SCHEMA = cv.Schema(
    {
        cv.GenerateID(CONF_MR60BHA2_ID): cv.use_id(MR60BHA2Component),
        cv.Optional(CONF_BREATH_RATE): sensor.sensor_schema(
            accuracy_decimals=0,
            state_class=STATE_CLASS_MEASUREMENT,
            icon=ICON_PULSE,
        ),
        cv.Optional(CONF_HEART_RATE): sensor.sensor_schema(
            accuracy_decimals=0,
            icon=ICON_HEART_PULSE,
            state_class=STATE_CLASS_MEASUREMENT,
            unit_of_measurement=UNIT_BEATS_PER_MINUTE,
        ),
        cv.Optional(CONF_DISTANCE): sensor.sensor_schema(
            device_class=DEVICE_CLASS_DISTANCE,
            state_class=STATE_CLASS_MEASUREMENT,
            unit_of_measurement=UNIT_CENTIMETER,
            accuracy_decimals=2,
            icon=ICON_SIGNAL,
        ),
        cv.Optional(CONF_NUM_TARGETS): sensor.sensor_schema(
            icon=ICON_COUNTER,
        ),
    }
)


async def to_code(config):
    mr60bha2_component = await cg.get_variable(config[CONF_MR60BHA2_ID])
    if breath_rate_config := config.get(CONF_BREATH_RATE):
        sens = await sensor.new_sensor(breath_rate_config)
        cg.add(mr60bha2_component.set_breath_rate_sensor(sens))
    if heart_rate_config := config.get(CONF_HEART_RATE):
        sens = await sensor.new_sensor(heart_rate_config)
        cg.add(mr60bha2_component.set_heart_rate_sensor(sens))
    if distance_config := config.get(CONF_DISTANCE):
        sens = await sensor.new_sensor(distance_config)
        cg.add(mr60bha2_component.set_distance_sensor(sens))
    if num_targets_config := config.get(CONF_NUM_TARGETS):
        sens = await sensor.new_sensor(num_targets_config)
        cg.add(mr60bha2_component.set_num_targets_sensor(sens))



"""DNS resolver for ESPHome using aioesphomeapi."""

from __future__ import annotations

import asyncio
import threading

from aioesphomeapi.core import ResolveAPIError, ResolveTimeoutAPIError
import aioesphomeapi.host_resolver as hr

from esphome.core import EsphomeError

RESOLVE_TIMEOUT = 10.0  # seconds


class AsyncResolver(threading.Thread):
    """Resolver using aioesphomeapi that runs in a thread for faster results.

    This resolver uses aioesphomeapi's async_resolve_host to handle DNS resolution,
    including proper .local domain fallback. Running in a thread allows us to get
    the result immediately without waiting for asyncio.run() to complete its
    cleanup cycle, which can take significant time.
    """

    def __init__(self, hosts: list[str], port: int) -> None:
        """Initialize the resolver."""
        super().__init__(daemon=True)
        self.hosts = hosts
        self.port = port
        self.result: list[hr.AddrInfo] | None = None
        self.exception: Exception | None = None
        self.event = threading.Event()

    async def _resolve(self) -> None:
        """Resolve hostnames to IP addresses."""
        try:
            self.result = await hr.async_resolve_host(
                self.hosts, self.port, timeout=RESOLVE_TIMEOUT
            )
        except Exception as e:  # pylint: disable=broad-except
            # We need to catch all exceptions to ensure the event is set
            # Otherwise the thread could hang forever
            self.exception = e
        finally:
            self.event.set()

    def run(self) -> None:
        """Run the DNS resolution."""
        asyncio.run(self._resolve())

    def resolve(self) -> list[hr.AddrInfo]:
        """Start the thread and wait for the result."""
        self.start()

        if not self.event.wait(
            timeout=RESOLVE_TIMEOUT + 1.0
        ):  # Give it 1 second more than the resolver timeout
            raise EsphomeError("Timeout resolving IP address")

        if exc := self.exception:
            if isinstance(exc, ResolveTimeoutAPIError):
                raise EsphomeError(f"Timeout resolving IP address: {exc}") from exc
            if isinstance(exc, ResolveAPIError):
                raise EsphomeError(f"Error resolving IP address: {exc}") from exc
            raise exc

        return self.result



# Dummy integration to allow relying on AsyncTCP
import esphome.codegen as cg
import esphome.config_validation as cv
from esphome.const import (
    PLATFORM_BK72XX,
    PLATFORM_ESP32,
    PLATFORM_ESP8266,
    PLATFORM_LN882X,
    PLATFORM_RTL87XX,
)
from esphome.core import CORE, CoroPriority, coroutine_with_priority

CODEOWNERS = ["@esphome/core"]

CONFIG_SCHEMA = cv.All(
    cv.Schema({}),
    cv.only_with_arduino,
    cv.only_on(
        [
            PLATFORM_ESP32,
            PLATFORM_ESP8266,
            PLATFORM_BK72XX,
            PLATFORM_LN882X,
            PLATFORM_RTL87XX,
        ]
    ),
)


@coroutine_with_priority(CoroPriority.NETWORK_TRANSPORT)
async def to_code(config):
    if CORE.is_esp32 or CORE.is_libretiny:
        # https://github.com/ESP32Async/AsyncTCP
        cg.add_library("ESP32Async/AsyncTCP", "3.4.5")
    elif CORE.is_esp8266:
        # https://github.com/ESP32Async/ESPAsyncTCP
        cg.add_library("ESP32Async/ESPAsyncTCP", "2.0.0")



from esphome import pins
import esphome.codegen as cg
import esphome.config_validation as cv
from esphome.const import CONF_CLOCK_PIN, CONF_DATA_PIN, CONF_ID

CODEOWNERS = ["@Cossid"]
MULTI_CONF = True

CONF_MAX_POWER_COLOR_CHANNELS = "max_power_color_channels"
CONF_MAX_POWER_WHITE_CHANNELS = "max_power_white_channels"

sm10bit_base_ns = cg.esphome_ns.namespace("sm10bit_base")
Sm10BitBase = sm10bit_base_ns.class_("Sm10BitBase", cg.Component)

SM10BIT_BASE_CONFIG_SCHEMA = cv.Schema(
    {
        cv.Required(CONF_DATA_PIN): pins.gpio_output_pin_schema,
        cv.Required(CONF_CLOCK_PIN): pins.gpio_output_pin_schema,
        cv.Optional(CONF_MAX_POWER_COLOR_CHANNELS, default=2): cv.int_range(
            min=0, max=15
        ),
        cv.Optional(CONF_MAX_POWER_WHITE_CHANNELS, default=4): cv.int_range(
            min=0, max=15
        ),
    }
).extend(cv.COMPONENT_SCHEMA)


async def register_sm10bit_base(config):
    var = cg.new_Pvariable(config[CONF_ID])
    await cg.register_component(var, config)

    data = await cg.gpio_pin_expression(config[CONF_DATA_PIN])
    cg.add(var.set_data_pin(data))
    clock = await cg.gpio_pin_expression(config[CONF_CLOCK_PIN])
    cg.add(var.set_clock_pin(clock))
    cg.add(var.set_max_power_color_channels(config[CONF_MAX_POWER_COLOR_CHANNELS]))
    cg.add(var.set_max_power_white_channels(config[CONF_MAX_POWER_WHITE_CHANNELS]))

    return var



import esphome.codegen as cg
from esphome.components import esp32_ble
from esphome.components.esp32 import add_idf_sdkconfig_option
from esphome.components.esp32_ble import CONF_BLE_ID
import esphome.config_validation as cv
from esphome.const import CONF_ID, CONF_TX_POWER, CONF_TYPE, CONF_UUID
from esphome.core import TimePeriod

AUTO_LOAD = ["esp32_ble"]
DEPENDENCIES = ["esp32"]

esp32_ble_beacon_ns = cg.esphome_ns.namespace("esp32_ble_beacon")
ESP32BLEBeacon = esp32_ble_beacon_ns.class_(
    "ESP32BLEBeacon",
    cg.Component,
    esp32_ble.GAPEventHandler,
    cg.Parented.template(esp32_ble.ESP32BLE),
)
CONF_MAJOR = "major"
CONF_MINOR = "minor"
CONF_MIN_INTERVAL = "min_interval"
CONF_MAX_INTERVAL = "max_interval"
CONF_MEASURED_POWER = "measured_power"


def validate_config(config):
    if config[CONF_MIN_INTERVAL] > config.get(CONF_MAX_INTERVAL):
        raise cv.Invalid("min_interval must be <= max_interval")
    return config


CONFIG_SCHEMA = cv.All(
    cv.Schema(
        {
            cv.GenerateID(): cv.declare_id(ESP32BLEBeacon),
            cv.GenerateID(CONF_BLE_ID): cv.use_id(esp32_ble.ESP32BLE),
            cv.Required(CONF_TYPE): cv.one_of("IBEACON", upper=True),
            cv.Required(CONF_UUID): cv.uuid,
            cv.Optional(CONF_MAJOR, default=10167): cv.uint16_t,
            cv.Optional(CONF_MINOR, default=61958): cv.uint16_t,
            cv.Optional(CONF_MIN_INTERVAL, default="100ms"): cv.All(
                cv.positive_time_period_milliseconds,
                cv.Range(
                    min=TimePeriod(milliseconds=20), max=TimePeriod(milliseconds=10240)
                ),
            ),
            cv.Optional(CONF_MAX_INTERVAL, default="100ms"): cv.All(
                cv.positive_time_period_milliseconds,
                cv.Range(
                    min=TimePeriod(milliseconds=20), max=TimePeriod(milliseconds=10240)
                ),
            ),
            cv.Optional(CONF_MEASURED_POWER, default=-59): cv.int_range(
                min=-128, max=0
            ),
            cv.Optional(CONF_TX_POWER, default="3dBm"): cv.All(
                cv.decibel, cv.enum(esp32_ble.TX_POWER_LEVELS, int=True)
            ),
        }
    ).extend(cv.COMPONENT_SCHEMA),
    validate_config,
)

FINAL_VALIDATE_SCHEMA = esp32_ble.validate_variant


async def to_code(config):
    cg.add_define("USE_ESP32_BLE_UUID")

    uuid = config[CONF_UUID].hex
    uuid_arr = [
        cg.RawExpression(f"0x{uuid[i : i + 2]}") for i in range(0, len(uuid), 2)
    ]
    var = cg.new_Pvariable(config[CONF_ID], uuid_arr)

    parent = await cg.get_variable(config[esp32_ble.CONF_BLE_ID])
    esp32_ble.register_gap_event_handler(parent, var)

    await cg.register_component(var, config)
    cg.add(var.set_major(config[CONF_MAJOR]))
    cg.add(var.set_minor(config[CONF_MINOR]))
    cg.add(var.set_min_interval(config[CONF_MIN_INTERVAL]))
    cg.add(var.set_max_interval(config[CONF_MAX_INTERVAL]))
    cg.add(var.set_measured_power(config[CONF_MEASURED_POWER]))
    cg.add(var.set_tx_power(config[CONF_TX_POWER]))

    cg.add_define("USE_ESP32_BLE_ADVERTISING")

    add_idf_sdkconfig_option("CONFIG_BT_ENABLED", True)
    add_idf_sdkconfig_option("CONFIG_BT_BLE_42_FEATURES_SUPPORTED", True)



import esphome.codegen as cg
from esphome.components import text_sensor
import esphome.config_validation as cv
from esphome.const import (
    CONF_DNS_ADDRESS,
    CONF_IP_ADDRESS,
    CONF_MAC_ADDRESS,
    ENTITY_CATEGORY_DIAGNOSTIC,
)

DEPENDENCIES = ["ethernet"]

ethernet_info_ns = cg.esphome_ns.namespace("ethernet_info")

IPAddressEthernetInfo = ethernet_info_ns.class_(
    "IPAddressEthernetInfo", text_sensor.TextSensor, cg.PollingComponent
)

DNSAddressEthernetInfo = ethernet_info_ns.class_(
    "DNSAddressEthernetInfo", text_sensor.TextSensor, cg.PollingComponent
)

MACAddressEthernetInfo = ethernet_info_ns.class_(
    "MACAddressEthernetInfo", text_sensor.TextSensor, cg.PollingComponent
)

CONFIG_SCHEMA = cv.Schema(
    {
        cv.Optional(CONF_IP_ADDRESS): text_sensor.text_sensor_schema(
            IPAddressEthernetInfo, entity_category=ENTITY_CATEGORY_DIAGNOSTIC
        )
        .extend(cv.polling_component_schema("1s"))
        .extend(
            {
                cv.Optional(f"address_{x}"): text_sensor.text_sensor_schema(
                    entity_category=ENTITY_CATEGORY_DIAGNOSTIC,
                )
                for x in range(5)
            }
        ),
        cv.Optional(CONF_DNS_ADDRESS): text_sensor.text_sensor_schema(
            DNSAddressEthernetInfo, entity_category=ENTITY_CATEGORY_DIAGNOSTIC
        ).extend(cv.polling_component_schema("1s")),
        cv.Optional(CONF_MAC_ADDRESS): text_sensor.text_sensor_schema(
            MACAddressEthernetInfo, entity_category=ENTITY_CATEGORY_DIAGNOSTIC
        ),
    }
)


async def to_code(config):
    if conf := config.get(CONF_IP_ADDRESS):
        ip_info = await text_sensor.new_text_sensor(config[CONF_IP_ADDRESS])
        await cg.register_component(ip_info, config[CONF_IP_ADDRESS])
        for x in range(5):
            if sensor_conf := conf.get(f"address_{x}"):
                sens = await text_sensor.new_text_sensor(sensor_conf)
                cg.add(ip_info.add_ip_sensors(x, sens))
    if conf := config.get(CONF_DNS_ADDRESS):
        dns_info = await text_sensor.new_text_sensor(config[CONF_DNS_ADDRESS])
        await cg.register_component(dns_info, config[CONF_DNS_ADDRESS])
    if conf := config.get(CONF_MAC_ADDRESS):
        mac_info = await text_sensor.new_text_sensor(config[CONF_MAC_ADDRESS])
        await cg.register_component(mac_info, config[CONF_MAC_ADDRESS])



import esphome.codegen as cg
from esphome.components import i2c, sensirion_common, sensor
import esphome.config_validation as cv
from esphome.const import (
    CONF_COMPENSATION,
    CONF_GAIN_FACTOR,
    CONF_ID,
    CONF_STORE_BASELINE,
    CONF_TEMPERATURE_SOURCE,
    DEVICE_CLASS_AQI,
    ICON_RADIATOR,
    STATE_CLASS_MEASUREMENT,
)

DEPENDENCIES = ["i2c"]
AUTO_LOAD = ["sensirion_common"]
CODEOWNERS = ["@SenexCrenshaw", "@martgras"]

sgp4x_ns = cg.esphome_ns.namespace("sgp4x")
SGP4xComponent = sgp4x_ns.class_(
    "SGP4xComponent",
    sensor.Sensor,
    cg.PollingComponent,
    sensirion_common.SensirionI2CDevice,
)

CONF_ALGORITHM_TUNING = "algorithm_tuning"
CONF_GATING_MAX_DURATION_MINUTES = "gating_max_duration_minutes"
CONF_HUMIDITY_SOURCE = "humidity_source"
CONF_INDEX_OFFSET = "index_offset"
CONF_LEARNING_TIME_GAIN_HOURS = "learning_time_gain_hours"
CONF_LEARNING_TIME_OFFSET_HOURS = "learning_time_offset_hours"
CONF_NOX = "nox"
CONF_STD_INITIAL = "std_initial"
CONF_VOC = "voc"
CONF_VOC_BASELINE = "voc_baseline"


def validate_sensors(config):
    if CONF_VOC not in config and CONF_NOX not in config:
        raise cv.Invalid(
            f"At least one sensor is required. Define {CONF_VOC} and/or {CONF_NOX}"
        )
    return config


GAS_SENSOR = cv.Schema(
    {
        cv.Optional(CONF_ALGORITHM_TUNING): cv.Schema(
            {
                cv.Optional(CONF_INDEX_OFFSET, default=100): cv.int_,
                cv.Optional(CONF_LEARNING_TIME_OFFSET_HOURS, default=12): cv.int_,
                cv.Optional(CONF_LEARNING_TIME_GAIN_HOURS, default=12): cv.int_,
                cv.Optional(CONF_GATING_MAX_DURATION_MINUTES, default=720): cv.int_,
                cv.Optional(CONF_STD_INITIAL, default=50): cv.int_,
                cv.Optional(CONF_GAIN_FACTOR, default=230): cv.int_,
            }
        )
    }
)

CONFIG_SCHEMA = cv.All(
    cv.Schema(
        {
            cv.GenerateID(): cv.declare_id(SGP4xComponent),
            cv.Optional(CONF_VOC): sensor.sensor_schema(
                icon=ICON_RADIATOR,
                accuracy_decimals=0,
                device_class=DEVICE_CLASS_AQI,
                state_class=STATE_CLASS_MEASUREMENT,
            ).extend(GAS_SENSOR),
            cv.Optional(CONF_NOX): sensor.sensor_schema(
                icon=ICON_RADIATOR,
                accuracy_decimals=0,
                device_class=DEVICE_CLASS_AQI,
                state_class=STATE_CLASS_MEASUREMENT,
            ).extend(GAS_SENSOR),
            cv.Optional(CONF_STORE_BASELINE, default=True): cv.boolean,
            cv.Optional(CONF_VOC_BASELINE): cv.hex_uint16_t,
            cv.Optional(CONF_COMPENSATION): cv.Schema(
                {
                    cv.Required(CONF_HUMIDITY_SOURCE): cv.use_id(sensor.Sensor),
                    cv.Required(CONF_TEMPERATURE_SOURCE): cv.use_id(sensor.Sensor),
                },
            ),
        }
    )
    .extend(cv.polling_component_schema("60s"))
    .extend(i2c.i2c_device_schema(0x59)),
    validate_sensors,
)


async def to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    await cg.register_component(var, config)
    await i2c.register_i2c_device(var, config)

    if CONF_COMPENSATION in config:
        compensation_config = config[CONF_COMPENSATION]
        sens = await cg.get_variable(compensation_config[CONF_HUMIDITY_SOURCE])
        cg.add(var.set_humidity_sensor(sens))
        sens = await cg.get_variable(compensation_config[CONF_TEMPERATURE_SOURCE])
        cg.add(var.set_temperature_sensor(sens))

    cg.add(var.set_store_baseline(config[CONF_STORE_BASELINE]))

    if CONF_VOC_BASELINE in config:
        cg.add(var.set_voc_baseline(CONF_VOC_BASELINE))

    if CONF_VOC in config:
        sens = await sensor.new_sensor(config[CONF_VOC])
        cg.add(var.set_voc_sensor(sens))
        if CONF_ALGORITHM_TUNING in config[CONF_VOC]:
            cfg = config[CONF_VOC][CONF_ALGORITHM_TUNING]
            cg.add(
                var.set_voc_algorithm_tuning(
                    cfg[CONF_INDEX_OFFSET],
                    cfg[CONF_LEARNING_TIME_OFFSET_HOURS],
                    cfg[CONF_LEARNING_TIME_GAIN_HOURS],
                    cfg[CONF_GATING_MAX_DURATION_MINUTES],
                    cfg[CONF_STD_INITIAL],
                    cfg[CONF_GAIN_FACTOR],
                )
            )

    if CONF_NOX in config:
        sens = await sensor.new_sensor(config[CONF_NOX])
        cg.add(var.set_nox_sensor(sens))
        if CONF_ALGORITHM_TUNING in config[CONF_NOX]:
            cfg = config[CONF_NOX][CONF_ALGORITHM_TUNING]
            cg.add(
                var.set_nox_algorithm_tuning(
                    cfg[CONF_INDEX_OFFSET],
                    cfg[CONF_LEARNING_TIME_OFFSET_HOURS],
                    cfg[CONF_LEARNING_TIME_GAIN_HOURS],
                    cfg[CONF_GATING_MAX_DURATION_MINUTES],
                    cfg[CONF_GAIN_FACTOR],
                )
            )
    cg.add_library(
        "Sensirion Gas Index Algorithm",
        None,
        "https://github.com/Sensirion/arduino-gas-index-algorithm.git#3.2.1",
    )
